import 'dart:convert';
import 'dart:io';
import 'package:connectivity/connectivity.dart';
import 'package:dio/adapter.dart';
import 'package:dio/dio.dart';
import 'package:dio_http_cache/dio_http_cache.dart';
import 'package:flutter_mylibrary/flutter_mylibrary.dart';
import 'package:flutter_mylibrary/http/exception/http_error.dart';
import 'package:flutter_mylibrary/http/http_data.dart';
import 'package:flutter_mylibrary/utils/utils.dart';
import 'package:pretty_dio_logger/pretty_dio_logger.dart';
import 'retry/dioconnectivity_request_retrier.dart';
import 'retry/retry_onconnectionchange_interceptor.dart';


/// 请求方法.
class Method {
  static final String get = "GET";
  static final String post = "POST";
  static final String put = "PUT";
  static final String head = "HEAD";
  static final String delete = "DELETE";
  static final String patch = "PATCH";
}


class _DioConfig {

  _DioConfig({
    this.baseUrl,
    this.connectTimeout,
    this.receiveTimeout,
    this.pem,
    this.pKCSPath,
    this.pKCSPwd,
    this.codeKey,
    this.msgKey,
    this.dataKey,
    this.successCode,
    this.responseType,
    this.prettyDioLogger,
    this.openLog = true,
    this.retry = true,
    this.retryInterceptor,
    this.headers,
    this.extra,
    this.contentType,
    this.validateStatus,
    this.followRedirects,
    this.onError,
    this.onFinish,
    this.onStart,
    this.enableCache = true,
    this.enableAutoCookie = false,
    this.interceptors
  });

  /// 服务器地址
  String baseUrl;

  /// 超时
  int connectTimeout;
  int receiveTimeout;

  /// 详细使用请查看dio官网 https://github.com/flutterchina/dio/blob/flutter/README-ZH.md#Https证书校验.
  /// PEM证书内容.
  String pem;
  /// 详细使用请查看dio官网 https://github.com/flutterchina/dio/blob/flutter/README-ZH.md#Https证书校验.
  /// PKCS12 证书路径.
  String pKCSPath;

  /// 详细使用请查看dio官网 https://github.com/flutterchina/dio/blob/flutter/README-ZH.md#Https证书校验.
  /// PKCS12 证书密码.
  String pKCSPwd;


  /// [int code]字段 key, 默认：code.
  String codeKey = 'code';

  ///[String msg]字段 key 默认：msg.
  String msgKey = 'msg';

  ///[T data]字段 key 默认：data.
  String dataKey = 'data';

  ///服务器成功返回的code
  String successCode = '0';
  String responseType;

  /// 打印拦截器 https://github.com/Milad-Akarie/pretty_dio_logger.
  PrettyDioLogger prettyDioLogger;
  bool openLog = true;

  /// 请求重试拦截器 https://github.com/lewcianci/dio_retry
  RetryInterceptor retryInterceptor;
  bool retry = true;



  Map<String, dynamic> headers;
  Map<String, dynamic> extra;
  String contentType;
  ValidateStatus validateStatus;
  bool followRedirects;
  Function onError;
  Function onFinish;
  Function onStart;
  bool enableCache = true;
  bool enableAutoCookie = false;
  List<Interceptor> interceptors;
}



class DioUtils {

  //服务器路径

  static DioUtils _instance = DioUtils._internal();

  Dio _dio;

  factory DioUtils() => _instance;

  _DioConfig _dioConfig;

  Function onError;

  Function onFinish;

  Function onStart;


  ///-----------------------------------------在这里自定义服务器返回的json字段-------------------------------------
  String _msgKey = 'message';
  String _codeKey = "code";
  String _dataKey = 'data';
  String _successCode = '0';



  ///-----------------------------------------工具类初始化设置-------------------------------------

  DioUtils._internal(){
    //基础配置
    _dio = new Dio(BaseOptions(
        contentType: Headers.formUrlEncodedContentType,
        connectTimeout: 5 * 1000,
        receiveTimeout: 30 * 1000
    ));
  }

  ///初始化公共属性
  ///
  /// [baseUrl] 地址前缀
  /// [connectTimeout] 连接超时赶时间
  /// [receiveTimeout] 接收超时赶时间
  /// [interceptors] 基础拦截器
  void init({String baseUrl,
    int connectTimeout,
    int receiveTimeout,
    String pem,
    String pKCSPath,
    String pKCSPwd,
    String codeKey,
    String msgKey,
    String dataKey,
    String successCode,
    String responseType,
    PrettyDioLogger prettyDioLogger,
    bool openLog = true,
    bool retry = true,
    RetryInterceptor retryInterceptor,
    Map<String, dynamic> headers,
    Map<String, dynamic> extra,
    String contentType,
    ValidateStatus validateStatus,
    bool followRedirects,
    Function onError,
    Function onFinish,
    Function onStart,
    bool enableCache = true,
    bool enableAutoCookie = false,
    List<Interceptor> interceptors}) {

    _dioConfig = _DioConfig(
        baseUrl: baseUrl,
        connectTimeout: connectTimeout,
        receiveTimeout: receiveTimeout,
        pem: pem,
        pKCSPath: pKCSPath,
        pKCSPwd: pKCSPwd,
        codeKey: codeKey,
        msgKey: msgKey,
        dataKey: dataKey,
        successCode: successCode,
        responseType: responseType,
        prettyDioLogger: prettyDioLogger,
        openLog: openLog,
        retry: retry,
        retryInterceptor: retryInterceptor,
        headers: headers,
        extra: extra,
        contentType: contentType,
        validateStatus: validateStatus,
        followRedirects: followRedirects,
        onError: onError,
        onFinish: onFinish,
        onStart: onStart,
        enableCache: enableCache,
        enableAutoCookie: enableAutoCookie,
        interceptors: interceptors
    );

    _dio.options = _dio.options.merge(
        baseUrl: baseUrl??_dio.options.baseUrl,
        connectTimeout: connectTimeout??_dio.options.connectTimeout,
        receiveTimeout: receiveTimeout??_dio.options.receiveTimeout,
        headers: Map.from(_dio.options.headers)..addAll(headers??{}),
        responseType: responseType??_dio.options.responseType,
        extra: Map.from(_dio.options.extra)..addAll(extra??{}),
        contentType: contentType??_dio.options.contentType,
        validateStatus: validateStatus??_dio.options.validateStatus,
        followRedirects: followRedirects??_dio.options.followRedirects
    );
    if (interceptors != null && interceptors.isNotEmpty) {
      _dio.interceptors..addAll(interceptors);
    }

    _codeKey = codeKey ?? _codeKey;
    _msgKey = msgKey ?? _msgKey;
    _dataKey = dataKey ?? _dataKey;
    _successCode = successCode ?? _successCode;
    this.onStart = onStart;
    this.onError = onError;
    this.onFinish = onFinish;

    if(openLog && isInDebugMode){
      _dio.interceptors.add(prettyDioLogger??PrettyDioLogger(
          requestHeader: true,
          requestBody: true,
          responseBody: true,
          responseHeader: false,
          error: true,
          compact: true,
          maxWidth: 90)
      );
    }

    if (retry) {
      _dio.interceptors.add(retryInterceptor??RetryInterceptor(
          dio: _dio,
          options: const RetryOptions(
            retries: 3, // Number of retries before a failure
            retryInterval: const Duration(seconds: 1), // Interval between each retry
          )
      ));
    }



    if(enableCache){
      _dio.interceptors.add(DioCacheManager(CacheConfig(baseUrl: baseUrl)).interceptor);
    }

    if(enableAutoCookie){
      getTemporaryDirectory().then((value){
        _dio.interceptors.add(CookieManager(PersistCookieJar(dir: value.path)));
      });
    }


    ///-----------------------------------------证书相关信息-------------------------------------
    if (_dio != null) {
      (_dio.httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate = (HttpClient client) {
        client.badCertificateCallback = (X509Certificate cert, String host, int port) {
          if (isNotEmpty(pem)) {
            if (cert.pem == pem) {
              // 证书一致，则放行
              return true;
            }else{
              return false;
            }
          }else{
            return true;
          }
        };
      };
      if (isNotEmpty(pKCSPath) && isNotEmpty(pKCSPwd)) {
        (_dio.httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate = (HttpClient client) {
          SecurityContext sc = new SecurityContext();
          //file为证书路径
          sc.setTrustedCertificates(pKCSPath, password: pKCSPwd);
          HttpClient httpClient = new HttpClient(context: sc);
          return httpClient;
        };
      }
    }
  }



  Dio get dio => _dio;

  ///-----------------------------------------请求方法封装-------------------------------------

  // ignore: missing_return
  Future<HttpData> request(
      String method,
      String path, {
        bool forceRefresh = true,
        bool noCache = false,
        String cacheKey,
        bool isFormData = false,
        bool publicMethod = true,
        data,
        Options options,
        CancelToken cancelToken,
        Function onStart,
        Function onError,
        Function onFinish,
        String successCode,
        String codeKey,
        String msgKey,
        String dataKey}) async{


    Response response;
    String _code;
    String _msg;
    var _resultData;
    HttpData _httpData;

    bool useFinish = false;
    try{
      if (onStart != null){
        onStart();
      }else if(this.onStart != null){
        if (publicMethod) {
          this.onStart();
        }
      }

      Options requestOptions = options ?? Options();

      if (isFormData) {
        data = FormData.fromMap(data);
      }

      if(!noCache){
        requestOptions = buildCacheOptions(
          Duration(days: 7),
          maxStale: Duration(days: 10),
          forceRefresh: forceRefresh,
          options: options,
        );
      }

      requestOptions.method = method;

      if (requestOptions.method == Method.get) {
        response = await DioUtils().dio.request(
            path,
            queryParameters: data,
            options: requestOptions,
            cancelToken: cancelToken
        );
      }else{
        response = await DioUtils().dio.request(
            path,
            data: data,
            options: requestOptions,
            cancelToken: cancelToken
        );
      }



      if (isNotEmpty(isNotEmpty(response.data))) {
        if (response.data is List) {
          _resultData = response.data;
          _code = _successCode;
        }else{
          Map<String, dynamic> resJson;
          if(response.data is Map){
            resJson = response.data;
          }else{
            resJson = json.decode(response.data);
          }
          _code = resJson[_codeKey].toString();
          _resultData = resJson[_dataKey];
          _msg = resJson[_msgKey];


          if(_code != _successCode){
            throw (HttpError(_code.toString(), _msg));
          }
        }

        _httpData = HttpData(
          code: _code,
          message: _msg,
          data: _resultData
        );

      }else{
        ///response.data为空
        throw(HttpError(response.statusCode.toString(), response.statusMessage));
      }


      if(!useFinish){
        if (isNotEmpty(onFinish)) {
          onFinish();
        } else if(this.onFinish != null){
          this.onFinish(response);
        }
        useFinish = true;
      }

      return _httpData;
    }catch(e, s){
      var error;
      if (e is DioError) {
        error = HttpError.dioError(e, s);
      }else{
        error = e;
      }

      if(onError != null){
        onError.call(error);
      }else if (this.onError != null) {
        if (publicMethod) {
          this.onError.call(error);
        }
      }
      if(!useFinish){
        if (isNotEmpty(onFinish)) {
          onFinish();
        } else if(this.onFinish != null){
          this.onFinish(response);
        }
        useFinish = true;
      }
      return Future.error(error);
    }
  }


  void setHeaders(Map<String, dynamic> headers){
    _dio.options.headers.addAll(headers);
  }

  void clearHeaders(Map<String, dynamic> headers){
    _dio.options.headers.clear();
  }

  ///修改默认url
  void changeBaseUrl(String baseUrl){
    _dio.options.baseUrl = baseUrl;
  }


  void fileUpload(String path, Map<String, dynamic> map, {Function onSendProgress, Function onStart, Function onError, Function onFinish}) async{
    FormData formData = FormData.fromMap(map);
    Response response;
    try{
      response = await _dio.post(
          path,
          data: formData,
          onSendProgress: (int sent, int total){
            if (onSendProgress != null) {
              onSendProgress.call(sent, total);
            }
          }
      );
    }on DioError catch(e, s){
      throw HttpError.dioError(e, s);
    }catch(e){
      if (e is HttpError) {
        if(onError != null){
          onError.call(e);
        }else if (this.onError != null) {
          this.onError.call(e);
        }

        if (isNotEmpty(onFinish)) {
          onFinish();
        } else if(this.onFinish != null){
          this.onFinish();
        }
      }
    }
  }



  ///-----------------------------------------dio部分设置-------------------------------------


  void enableRetry(){
    _dio.interceptors.add(
      RetryOnConnectionChangeInterceptor(
        requestRetrier: DioConnectivityRequestRetrier(
          dio: _dio,
          connectivity: Connectivity(),
        ),
      ),
    );
  }


  void setCookie(String cookie){
    Map<String, dynamic> _headers = new Map();
    _headers["Cookie"] = cookie;
    _dio.options.headers.addAll(_headers);
  }

}
